/* Implement a portal engine core for scenery renderization.

Has the main visualization pipeline: make transformations from universe to
camera reference system, make poly clip, draw polygon mesh on frame buffer,
call recursively functions to draw other visible objects and sectors.

Portal is a polygon where other polygons are clipped, used to define visible
sectors from where the camera be.

Please refer doc/notes.txt for more info about the view pipeline and
engine details.

The main function is mRenderMesh. It call:
   For each mesh
      mApplyCamera - transform SRU -> SRC (and scale mesh)
      mAjust - project vertexes
      For each poly:
         mDefinePolygonData - set poly vertexes, light info,...
         mPassBackFaceCulling - try back face culling
         mGetPolyTextureData - get texture for the poly
         mMakeZNearPlaneClip  - clip poly to near Z
and draw a poly mesh.

WARNING:   
- polygons vertexes must be in anticlockwise order
*/



#include <string.h>

#include <math.h>

#include "../defs.h"

#include "../mem.h"

#include "../array.h"

#include "../string.h"

#include "../texture.h"

#include "../file.h"

#include "../filerem.h"

#include "../filepcx.h"

#include "../lighttab.h"

#include "../draw/draw.h"

#include "../node.h"

#include "../nodelii.h"

#include "../list.h"

#include "../polygon.h"

#include "../nodeflo.h"

#include "../nodepol.h"

#include "../nodexyz.h"

#include "../nodexy.h"

#include "../map.h"

#include "../math.h"

#include "../mesh.h"

#include "../sector.h"

#include "../nodesec.h"

#include "../refsys.h"

#include "../camsys.h"

#include "../object.h"

#include "../nodeint.h"

#include "../nodetex.h"

#include "../nodeobj.h"

#include "../nodelgt.h"

#include "../filetxe.h"

#include "../eventtab.h"

#include "../scenery.h"

#include "render.h"

#include "../comp/nodecomp.h"

#include "../comp/cokbcont.h"

#include "../comp/codata.h"

#include "../comp/cocamsys.h"

#include "../comp/comesh.h"



/* constructor - init empty
**********************************************************/
CRender::CRender ()
{
   m_pastPolVertexCoords = NULL;
   m_dwMaxVertArray = 0;
   m_pastPolZClipTex = m_pastPolZClipCoords = NULL;
   #ifdef COMP_RENDER_INTERPOLATED_LIGHT
   m_pafPolVertexLights = NULL;
   m_pafPolZClipLight = NULL;
   #endif
   m_dwMaxZClipArray = 0;

   mDestroy ();
}



/* constructor - init fly system
**********************************************************/
CRender::CRender ( CScenery& oScen )
{
   m_pastPolVertexCoords = NULL;
   m_dwMaxVertArray = 0;
   m_pastPolZClipTex = m_pastPolZClipCoords = NULL;
   #ifdef COMP_RENDER_INTERPOLATED_LIGHT
   m_pafPolVertexLights = NULL;
   m_pafPolZClipLight = NULL;
   #endif
   m_dwMaxZClipArray = 0;

   mCreate ( oScen );
}



/* destructor
**********************************************************/
CRender::~CRender ()
{
   mDestroy ();
}



/* Init render engine.
oScen define the scenery to render
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CRender::mCreate ( CScenery& oScen )
{
   mDestroy ();

   m_poScen = &oScen;      // define scenery pointer

   // REVIEW
   // define texture vertexes, used in textures of size 256x256
   // polys with 3 vertexes
   m_astTexVertexes3 [ 0 ].nX = 0;
   m_astTexVertexes3 [ 0 ].nY = 0;
   m_astTexVertexes3 [ 1 ].nX = 0;
   m_astTexVertexes3 [ 1 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes3 [ 2 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes3 [ 2 ].nY = 0;
   // polys with 4 vertexes
   m_astTexVertexes4 [ 0 ].nX = 0;
   m_astTexVertexes4 [ 0 ].nY = 0;
   m_astTexVertexes4 [ 1 ].nX = 0;
   m_astTexVertexes4 [ 1 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes4 [ 2 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes4 [ 2 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes4 [ 3 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes4 [ 3 ].nY = 0;
   // polys with 5 vertexes
   m_astTexVertexes5 [ 0 ].nX = 0;
   m_astTexVertexes5 [ 0 ].nY = 0;
   m_astTexVertexes5 [ 1 ].nX = 0;
   m_astTexVertexes5 [ 1 ].nY = (INT) ( (float) TEXTURE_MAXDIM * 3 / 4 );
   m_astTexVertexes5 [ 2 ].nX = TEXTURE_MAXDIM / 2;
   m_astTexVertexes5 [ 2 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes5 [ 3 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes5 [ 3 ].nY = (INT) ( (float) TEXTURE_MAXDIM * 3 / 4 );
   m_astTexVertexes5 [ 4 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes5 [ 4 ].nY = 0;
   // polys with 6 vertexes
   m_astTexVertexes6 [ 0 ].nX = (INT) ( (float) TEXTURE_MAXDIM * 1 / 4 );
   m_astTexVertexes6 [ 0 ].nY = 0;
   m_astTexVertexes6 [ 1 ].nX = 0;
   m_astTexVertexes6 [ 1 ].nY = TEXTURE_MAXDIM / 2;
   m_astTexVertexes6 [ 2 ].nX = (INT) ( (float) TEXTURE_MAXDIM * 1 / 4 );
   m_astTexVertexes6 [ 2 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes6 [ 3 ].nX = (INT) ( (float) TEXTURE_MAXDIM * 3 / 4 );
   m_astTexVertexes6 [ 3 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes6 [ 4 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes6 [ 4 ].nY = TEXTURE_MAXDIM / 2;
   m_astTexVertexes6 [ 5 ].nX = (INT) ( (float) TEXTURE_MAXDIM * 3 / 4 );
   m_astTexVertexes6 [ 5 ].nY = 0;
   // polys with 7 vertexes (not very precise)
   m_astTexVertexes7 [ 0 ].nX = TEXTURE_MAXDIM / 2;
   m_astTexVertexes7 [ 0 ].nY = 0;
   m_astTexVertexes7 [ 1 ].nX = 0;
   m_astTexVertexes7 [ 1 ].nY = TEXTURE_MAXDIM / 3;
   m_astTexVertexes7 [ 2 ].nX = 0;
   m_astTexVertexes7 [ 2 ].nY = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes7 [ 3 ].nX = TEXTURE_MAXDIM / 3;
   m_astTexVertexes7 [ 3 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes7 [ 4 ].nX = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes7 [ 4 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes7 [ 5 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes7 [ 5 ].nY = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes7 [ 6 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes7 [ 6 ].nY = TEXTURE_MAXDIM / 3;
   // polys with 8 vertexes
   m_astTexVertexes8 [ 0 ].nX = TEXTURE_MAXDIM / 3;
   m_astTexVertexes8 [ 0 ].nY = 0;
   m_astTexVertexes8 [ 1 ].nX = 0;
   m_astTexVertexes8 [ 1 ].nY = TEXTURE_MAXDIM / 3;
   m_astTexVertexes8 [ 2 ].nX = 0;
   m_astTexVertexes8 [ 2 ].nY = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes8 [ 3 ].nX = TEXTURE_MAXDIM / 3;
   m_astTexVertexes8 [ 3 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes8 [ 4 ].nX = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes8 [ 4 ].nY = TEXTURE_MAXDIM;
   m_astTexVertexes8 [ 5 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes8 [ 5 ].nY = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes8 [ 6 ].nX = TEXTURE_MAXDIM;
   m_astTexVertexes8 [ 6 ].nY = TEXTURE_MAXDIM / 3;
   m_astTexVertexes8 [ 7 ].nX = (INT) ( (float) TEXTURE_MAXDIM / 1.5 );
   m_astTexVertexes8 [ 7 ].nY = 0;

   return m_oDraw.mCreate ();
}



/* Destroy: make pointers point to NULL, free memory.
**********************************************************/
void CRender::mDestroy ()
{
   m_poScen = NULL;

   m_dwMaxVertArray = 0;
   m_dwMaxZClipArray = 0;

   // free array memory
   if ( m_pastPolVertexCoords != NULL ) {
      delete m_pastPolVertexCoords;
      m_pastPolVertexCoords = NULL;
   }

   // free array memory for clipped vertexes
   if ( m_pastPolZClipCoords != NULL ) {
      delete m_pastPolZClipCoords;
      m_pastPolZClipCoords = NULL;
   }

   // free array memory for clipped vertexes
   if ( m_pastPolZClipTex != NULL ) {
      delete m_pastPolZClipTex;
      m_pastPolZClipTex = NULL;
   }

   #ifdef COMP_RENDER_INTERPOLATED_LIGHT

   // free array with light levels
   if ( m_pafPolVertexLights != NULL ) {
      delete m_pafPolVertexLights;
      m_pafPolVertexLights = NULL;
   }

   // free array memory for clipped light information
   if ( m_pafPolZClipLight != NULL ) {
      delete m_pafPolZClipLight;
      m_pafPolZClipLight = NULL;
   }

   #endif

   m_oCamVert.mRemoveAll ();
   m_oPrjVert.mRemoveAll ();

   m_oDraw.mDestroy ();
}



/* Set a frame buffer
**********************************************************/
BOOL CRender::mSetFrameBuffer ( INT nResX, INT nResY, CMem& oFrBuff )
{
   // verify Y resolution limitation
   if ( nResY > MAX_RESY )
      ABORT ( MSG_ERROR_OVER_RESY_LIMIT, __FILE__, __LINE__ )

   return m_oDraw.mSetFrameBuffer ( nResX, nResY, oFrBuff );
}
